/************************************************************************
* Зарядка Деда Ивана со Скифа))
* Charger_one.ino
* Created: 10/19/2014 7:30:46 PM
* Author: SunSB, 2014
*/ 


//#define DEBUG   1  // отладочные сообщения в порт
//#define DEBUG_PROGRAM   1  // отладочные сообщения в порт
//#define DEBUG_DISABLE_DATALOG 1 // запрет вывода данных для arduinoReader

// Phase-Frequency Correct mode (WGM1 = 8) runs at half-speed (it counts both up and down in each cycle), so the number of ticks is half what you’d expect. You could use Fast PWM mode or anything else, as long as you get the counts right; see page 133 of the Fine Manual.
// Phase-freq Correct timer mode -> runs at half the usual rate
#define PERIOD_TICKS(ticks) ( microsecondsToClockCycles( ticks / 2) / 1 )
#define getUnixTime()       (millis()/1000)
#ifndef cbi       // при FASTADC 1 analogRead работает 13,5мкс (1 такт АЦП=1/16000000*prescaler = 62.5ns*16=1000ns), в обычном состоянии 108мкс
#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#endif
#ifndef sbi
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))
#endif

#include <avr/pgmspace.h>
#include <EEPROM.h>
#include <LiquidCrystal.h>

#ifndef CONST_VAR_FOLDING
const double SENSORS_MAX[6]={ // максимумы значений для датчиков на ногах АЦП
    0.0,  // кнопки экрана
    16.0, // напряжение на АКБ
    40.0, // импульсное напряжение на АКБ
    50.0, // DC ток заряда или разряда
    0.0,
    0.0
};
const int16_t DC_CURRENT_SENSOR_ERROR=110; // ошибка нуля датчика тока в миллиамперах

const double KOEFF_MAIN_COIL    = 17.0/100.0;   // коэфф нарастания тока в конкретной железяке катушке   17.0/100.0 -> 17A за 50мкс, типа индуктивность
const double KOEFF_CHARGE_PWM   =0.05;          // на сколько увеличивается ток при увеличении PWM ( уменьшение скважности )
const double KOEFF_DISCHARGE_PWM=0.05;
const double ONE_DIV_1024  = 1.0/1024;          // там умножать а не делить будем
const uint16_t RELAY_WAIT  =     1000;          // в миллисекундах ожидание времени срабатывания реле

const uint8_t ADDR_BASE = 20; // адреса в EEPROM
const uint8_t ADDR_PULSE_LENGTH_0 = (ADDR_BASE+0); // word
const uint8_t ADDR_PULSE_LENGTH_1 = (ADDR_BASE+1); // 
const uint8_t ADDR_PULSE_PERIOD_0 = (ADDR_BASE+2); // word
const uint8_t ADDR_PULSE_PERIOD_1 = (ADDR_BASE+3); // 
const uint8_t ADDR_statePulse     = (ADDR_BASE+4); // byte  
const uint8_t ADDR_stateCurrent   = (ADDR_BASE+5); // byte
const uint8_t ADDR_currentPWM     = (ADDR_BASE+6); // byte

const uint8_t ADDR_EEPROGRAM_LINES_COUNT  =  99;
const uint8_t ADDR_EEPROGRAM        = 100;          // начало программы
// ноги
const uint8_t PIN_PULSE           = 10; // импульс 
const uint8_t PIN_POWER_ON        =  2; // включение силовой части
const uint8_t PIN_PULSE_UP_DN     = 12; // импульс заряжает 1 / разряжает 0, перебрасывает реле на первичной катушке

const uint8_t PIN_CHA_DIS_PWM     =  3; // шим зарядки / разрядки
const uint8_t PIN_DC_POWER_ON     = 11; // включение силовой части постоянного тока
const uint8_t PIN_CURRENT_CHA_DIS = 13; // флаг подключения зарядки 0 / разрядки 1 постоянным током

/*const uint8_t PIN_LCD_KEYS        = A0; // на плате распаян A0
const uint8_t PIN_VOLTAGE         = A1; // напряжение на батарее
const uint8_t PIN_VOLTAGE_PULSE   = A2; // напряжение на батарее импульсное
const uint8_t PIN_CURRENT         = A3; // постоянный ток зарядки или разрядки от режима
*/
// ==================================
const uint8_t ADMUX_LCD_KEYS      = 64; // // на плате распаян A0
const uint8_t ADMUX_VOLTAGE       = 65; // A1 напряжение на батарее
const uint8_t ADMUX_VOLTAGE_PULSE = 66; // A2 напряжение на батарее импульсное
const uint8_t ADMUX_CURRENT       = 67; // A3 постоянный ток зарядки или разрядки от режима
const uint8_t ADMUX_A4 = 68;  // A4
const uint8_t ADMUX_A5 = 69;  // A5

/*
const uint8_t ADMUX_A0=64; // A0=14
const uint8_t ADMUX_A1=65; // A1=15
const uint8_t ADMUX_A2=66; // A2=16
const uint8_t ADMUX_A3=67; // A3=17
const uint8_t ADMUX_A4=68; // A4=18
const uint8_t ADMUX_A5=69; // A5=19
*/

const String logM="m:";

#ifndef SYMBOLS_FOLDING // битовые определения символов для экрана 
const uint8_t SYMBOL_HEIGHT=8;
byte arrow_up[SYMBOL_HEIGHT] =
{
    B00100,
    B01110,
    B10101,
    B00100,
    B00100,
    B00100,
    B00100,
    B00100
};
byte arrow_dn[SYMBOL_HEIGHT] =
{
    B00100,
    B00100,
    B00100,
    B00100,
    B00100,
    B10101,
    B01110,
    B00100,
};
byte scul[SYMBOL_HEIGHT] =
{
    B01110,
    B10101,
    B11011,
    B01110,
    B00000,
    B10001,
    B01110,
    B10001,
};
byte happy[SYMBOL_HEIGHT] =
{
    B01110,
    B11111,
    B10101,
    B11111,
    B10101,
    B11011,
    B01110,
    B00000,
};
const char ch_arrow_up=1;
const char ch_arrow_dn=2;
const char ch_scul    =3;
const char ch_happy   =4;
const char ch_arrow_rt=126;
const char ch_arrow_lt=127;
#endif

#ifdef __PGMSPACE_H_ // сам ifdef нах не нужен, но редактор может на нем collapse сделать
// PROGMEM
// эти финты ушами чтоб строки сохранились не в оперативной памяти, а в памяти программы
//                                   ================ 16 символов
prog_char str_CALIBR[]     PROGMEM ="Calibrate";
prog_char str_PULSE[]      PROGMEM ="Puls ";
prog_char str_DC[]         PROGMEM ="DC ";
prog_char str_HOLDS[]      PROGMEM ="Hold";
prog_char str_STAGE[]      PROGMEM ="Stage";
prog_char str_PROG[]       PROGMEM ="Prog";
PROGMEM const char* commands[]={ str_CALIBR, str_PULSE, str_DC, str_HOLDS, str_STAGE, str_PROG};
//                                      ================ 16 символов
prog_char str_STARTING     [] PROGMEM ="Starting";
prog_char str_CANCEL       [] PROGMEM ="Cancel ";
prog_char str_CHARGE       [] PROGMEM ="charge";
prog_char str_DISCHARGE    [] PROGMEM ="discharge";
prog_char str_NONE         [] PROGMEM ="none";
prog_char str_PERIOD       [] PROGMEM ="Period";
prog_char str_PWM          [] PROGMEM ="PWM";
prog_char str_WAIT         [] PROGMEM ="Wait ";
prog_char str_V            [] PROGMEM ="V ";
prog_char str_A            [] PROGMEM ="A ";
prog_char str_OK           [] PROGMEM ="ok";
//                                       ================ 16 символов
prog_char str_HOLD_AMPS     [] PROGMEM ="DC amps";
prog_char str_HOLD_MIN_VOLTS[] PROGMEM ="Min AKB";
prog_char str_HOLD_MAX_VOLTS[] PROGMEM ="Max AKB";
//                                  ================ 16 символов

const char char_SEMI=':'; 
const char char_SPACE=' ';
#endif

#ifndef COMMANDS_PART // сам ifdef нах не нужен, но редактор может на нем collapse сделать
// привязываем к *headers[] для экономии
const byte VIEW__CALIBRATOR   = 0;
const byte VIEW__PULSE        = 1;
const byte VIEW__CURRENT      = 2;
const byte VIEW__HOLDS        = 3;
const byte VIEW__STAGE        = 4;
const byte VIEW__PROGRAMM     = 5;
const byte VIEW__LAST_CMD     = VIEW__PROGRAMM;
int8_t nowOnScreen=0;

// текущий режим работы
const uint8_t ST__STAGE_CHARGE   = B00001; // зарядка, вывод по чем делать? по прикидкам +импульс ток -пост?
const uint8_t ST__STAGE_DISCHARGE= B00010; // разрядка, вывод по чем делать?
const uint8_t ST__PROGRAMM       = B00100; // афтоматическое массажирование поциента ))
const uint8_t ST__STAGE_WAIT     = B01000; // ожидание - отстой ))
uint8_t state=0;              // ничего не происходит

// текущий режим работы импульсной части
const byte ST__PULSE_UP  = 1; // зарядка
const byte ST__PULSE_DN  = 2; // разрядка

// текущий режим работы з/р постоянным током
const byte ST__DC_CHARGE  = 1; // зарядка
const byte ST__DC_DISCHA  = 2; // разрядка
// страницы настройки импульса и ШИМ
const uint8_t PG_PULSE_DIR       =0; // выбирать selectом, цикл в одну сторону
const uint8_t PG_PULSE_WIDTH     =1;
const uint8_t PG_PULSE_PERIOD    =2;
uint8_t pagePulse=0;

const uint8_t PG_DC_DIR =0;
const uint8_t PG_DC_PWM =1;
uint8_t pageCurrent=0;

const uint8_t PG_HOLD_AMPS       =0;
const uint8_t PG_HOLD_MIN_VOLTS  =1;
const uint8_t PG_HOLD_MAX_VOLTS  =2;
uint8_t pageHolds=0;
#endif


const uint32_t LOG_WRITE_FREQUENCY    = 10*1000; // частота записей в лог в миллисекундах
const uint32_t logHeaderWriteFrequency= 10*LOG_WRITE_FREQUENCY; // частота записей заголовка в лог в миллисекундах
                                                                // на случай неприятностей, не используется
uint32_t lastLogWriting=0;//, lastLogHeaderWriting=0;
                                                   
const uint8_t OSCILL_POINTS=10; // сколько измерений для осцилла делать
const uint8_t OSCILL_START = 4; // индекс первай точки осцила в sensorsData
// индексы в массиве sensorsData
// такой финт позволит очень быстро чистить одной командой
const uint8_t SAMPLE_NUM  =0; // - номер  текущего loop()
const uint8_t SENS_VOLTAGE=1; // - замер напряжения перед импульсом с PIN_VOLTAGE для лога
const uint8_t SENS_CURRENT=2; // - замер тока с PIN_CURRENT для лога
const uint8_t OSCILL_TIME =3; // - время затраченное на все замеры осцилла
// дальше будет OSCILL_POINTS замеров для осцилла с PIN_VOLATAGE_PULSE
uint32_t  sensorsData[4+OSCILL_POINTS]={0};
uint8_t   OSCILL_INDEX = OSCILL_START+OSCILL_POINTS; // индекс текущей точки для записи данных осцилла

uint32_t  sensorsDataLCD[3]={0}; // тоже самое для LCD
    
double nowV=15, nowI=0, nowV_LCD=0; // текущие напряжение и ток на АКБ


uint32_t unixTimeStamp=0, lastLCDshowTime=0;//, lastStageW_LCD=0;
// для подсчёта ёмкости и энергии
uint32_t lastSampleTime=0; double   prevSampleWatt=0, prevSampleAmps=0;
#ifndef KBD_FOLDING
const uint8_t KEY_RIGHT  =0;  // LCD
const uint8_t KEY_UP     =1;
const uint8_t KEY_DOWN   =2;
const uint8_t KEY_LEFT   =3;
const uint8_t KEY_SELECT =4;
const uint8_t NUM_KEYS   =5;
const uint16_t adc_key_val[5] ={30, 150, 360, 535, 760 };
byte key=NUM_KEYS, oldKey=NUM_KEYS, oldHandledKey=NUM_KEYS, keySelectCount=0;
uint32_t keyPressMills=0, keyPressDiff=1000, keyPressRepeat=0, keyPressUnixTime=0; 
#endif

uint16_t pulseLengthInTicks=0;   // перерассчитывается при изменении, для определения импульс / пауза
//uint32_t stagePulseCount=0;      // счетчик импульсов, по нему можно посчитать чегонить. наверное. потом.
int8_t   stageProgramDisplayX=0; // для отображения текста стадии на LCD


struct  STAGE{
    char name;
    uint8_t  progLine;
    uint8_t  nextProgLine;    // номер строки программы который будет запрошен следующим
    uint32_t time;            // время выполнения стадии в секундах. 32 потому как в нем нужно сохранять unixTimeStamp старта
    double   watts;           // сколько энергии накачало
    double   amps;            // сколько амперов накачало

    uint8_t  statePulse;      // состояние импульсов, заряд/разряд
    uint16_t pulseLength;     // длительность импульса в микросекундах
    uint16_t pulsePeriod;     // период импульса в микросекундах

    uint8_t  stateCurrent;    // состояние постоянного тока стадии, заряд/разряд
    uint8_t  currentPWM;      // период импульса в микросекундах
    double   holdAmps;        // удерживать ток заряда/разряда
    uint16_t waitTime;        // сколько ждать, хранится в строке команды 
    uint32_t waitingFor;      // waitingFor=getUnixTime()+waitTime, ждём до getUnixTime() == stage.WaitingFor
    
//    double checkHigherThen; // стадия проверки напряжения на больше чем
//    double checkLowerThen;  // стадия проверки напряжения на меньше чем
    double maxVoltage;        // максимальное напряжение достигнутое в пределах стадии
    double minVoltage;        // минимальное напряжение достигнутое в пределах стадии
    double tg;                // изменение напряжения за некий период ?????
    uint16_t tgTime;          // интервал в секундах через который проверяется
    uint16_t tgTimeCounter;   // для измерений
    double   prevTgVoltage;   // для измерений
    double   tgVoltage;   // для измерений
    
};
STAGE stage={0};
const uint8_t MAX_PROG_LINES =20;

struct PROG_LINE_STATS{
    uint16_t execCount; // сколько раз выполниласть
    uint32_t time;      // сколько времени выполнялась
    double   watts;     // сколько мощности прокачала
    double   amps;     // сколько мощности прокачала
};
struct PROGRAM{
    bool programming;                       // флаг получения программы
    bool debug;                             // флаг отладки программы
    bool debugTrue;                         // при отладке на все вопросы true
    uint8_t progLinesCount;                    // к-во строк программы в EEPROM
    uint32_t time;                          // начало выполнения
    int8_t   stageIndex;                    // индекс текущей строки или -1 при ошибке
    String   stageProgram;                  // строка стадии в пвмяти

    double   watts;                         // итог работы в ваттах 
    double   amps;                          // итог работы в амперах

    uint16_t eeStages[MAX_PROG_LINES];      // адреса строк стадий в EEPROM
    int16_t  indexToLine[MAX_PROG_LINES];   // соответствие индексов строк к их номерам в программе
    PROG_LINE_STATS stats[MAX_PROG_LINES];  // статистика выполнения
};
PROGRAM program={
    programming   : false,
    debug         : false,
    debugTrue     : false, 
    progLinesCount: 0,
    time          : 0,
    stageIndex    :-1,
    stageProgram  : "",
    watts         : 0,
    amps          : 0,
    eeStages:{0},
    indexToLine:{0},
    stats:{0}
};

String serialString="";// сюда читается из порта
#endif

LiquidCrystal lcd(8, 9, 4, 5, 6, 7); // инициализация экрана

/*  ======================================= */
#ifndef UTILS_FOLDING
String toHours( int32_t secs, char sep ){
    String s;
    if( secs>3600 )
        s+=int(secs/3600);
    s+=sep;
    secs = secs%3600;
    if( s.length()>1 || secs>60 ) 
        s+= int(secs/60);
    s+=sep;
    s+=secs%60;
    return s;
}
   
String progMemToString( PGM_P arr[], byte index ){
    char ch[40];
    strcpy_P( ch, (char*)pgm_read_word( &(arr[index]) )); //select mode
    return String( ch );
}
String progMemToString(PGM_P str) {
    // считывает строку данных из флеш памяти программы в оперативную память ардуины
    // возвращает объект String
    String s="";
    for (uint8_t c; (c = pgm_read_byte(str)); str++) s+=(char)c;
    return s;
}
String intToString(long val, int digits=4){
    // форматирует целое число в строку с frac знаками сдвигает вправо
    String s = String( int(val) );
    digits--;
    while( digits>0){
        val/=10;
        if( 0==val ){
            s= char_SPACE+s;
        }            
        digits--;
    }
    return s;
}
String doubleToString(double val, uint8_t digits=0, uint8_t frac=2){
    // форматирует вещественное число в строку с frac знаками после точки
    //    val += 5.0/mul/10; // округление вместо обрезания ))
    String s;
    double v=0.5;
    for( uint8_t i=frac; i>0; i--) v/=10;
    val+=v; // округляем
    v=abs(val);
    if( digits )
        s = intToString( long(v), digits );
    else
        s = String(int(v));
    if( frac>0 ){
        s+='.';
        while( frac>0 ){
            v-=int(v);
            v*=10;
            s+= String(int(v));
            frac--;
        }
    }        
    if(val<0) s='-'+s;
    return s;
}
String doubleToStringStepped( double val ){
    if( abs(val) < 1 ) return doubleToString( val, 0, 3);
    else
    if( abs(val) < 10 ) return doubleToString( val, 0, 2);
    else
    if( abs(val) < 100 ) return doubleToString( val, 0, 1);
    else
    return doubleToString( val, 0, 0);
}
int freeRam () { // определяет сколько свободной памяти
    extern int __heap_start, *__brkval;
    int v;
    return (int) &v - (__brkval == 0 ? (int) &__heap_start : (int) __brkval);
}
uint16_t eeWordRead ( uint16_t addr ){
    uint16_t w=EEPROM.read( addr+1 );
    w<<=8;
    w+=EEPROM.read( addr );
    return w;
}
void   eeWordWrite  ( uint16_t addr, uint16_t w ){
    EEPROM.write( addr, (uint8_t)w );
    w>>=8;
    EEPROM.write( addr+1, (uint8_t)w );
}
void   eeStringWrite( uint16_t addr, String s ){
    EEPROM.write( addr, s.length() ); // вписываем длинну строки
    for( uint8_t i=0; i<s.length(); i++ ){
        EEPROM.write( addr+1+i, s.charAt(i) );
    }
}
String eeStringRead ( uint16_t addr ){
    String s="";
    uint8_t length=EEPROM.read(addr);
    if( length==0xFF ){
        if(EEPROM.read(addr+1)==0xFF) return s; // еще ничего не писалось
    }
    addr++;
    for( int i=0; i<length; i++ ){
        s+=(char)EEPROM.read( addr+i );
    }
    return s;
}

void eeAllRead(){
    stage.pulseLength = eeWordRead ( ADDR_PULSE_LENGTH_0 );
    stage.pulsePeriod = eeWordRead ( ADDR_PULSE_PERIOD_0 );
    stage.statePulse  = EEPROM.read( ADDR_statePulse );
    stage.stateCurrent= EEPROM.read( ADDR_stateCurrent );
    stage.currentPWM  = EEPROM.read( ADDR_currentPWM );
}
void eeAllWrite(){
    eeWordWrite ( ADDR_PULSE_LENGTH_0    , stage.pulseLength );
    eeWordWrite ( ADDR_PULSE_PERIOD_0    , stage.pulsePeriod );
    EEPROM.write( ADDR_statePulse        , stage.statePulse  );
    EEPROM.write( ADDR_stateCurrent      , stage.stateCurrent);
    EEPROM.write( ADDR_currentPWM        , stage.currentPWM  );
}
void setPwmFrequency(int pin, int divisor) {
    byte mode;
    if(pin == 5 || pin == 6 || pin == 9 || pin == 10) {
        switch(divisor) {
            case 1: mode = 0x01; break;
            case 8: mode = 0x02; break;
            case 64: mode = 0x03; break;
            case 256: mode = 0x04; break;
            case 1024: mode = 0x05; break;
            default: return;
        }
        if(pin == 5 || pin == 6) {
            TCCR0B = (TCCR0B & 0b11111000) | mode;
        } else {
            TCCR1B = (TCCR1B & 0b11111000) | mode;
        }
    } else if(pin == 3 || pin == 11) {
        switch(divisor) {
            case 1: mode = 0x01; break;
            case 8: mode = 0x02; break;
            case 32: mode = 0x03; break;
            case 64: mode = 0x04; break;
            case 128: mode = 0x05; break;
            case 256: mode = 0x06; break;
            case 1024: mode = 0x7; break;
            default: return;
        }
        TCCR2B = (TCCR2B & 0b11111000) | mode;
    }
}
// ========================
void initPulsePWM(uint16_t period_uSec, uint16_t pulse_uSec ){
    TCCR1B = 0x00; // stop Timer1 clock for register updates
    // Clear OC1A on match, P-F Corr
    // PWM Mode: lower WGM1x = 00
    TCCR1A = 0x80 | 0x00;
    TCCR1A|= 0x20;
    ICR1   = PERIOD_TICKS(period_uSec);
    OCR1B  = pulseLengthInTicks = PERIOD_TICKS(pulse_uSec);
    TCCR1B = 0x10 | 0x01;  // CS2:0 bits = prescaler selection;
}
// запуск АЦП, через какоето время данные дождёмся и считаем
void myAnalogReadStart(uint8_t ADMUX_VAL){
//    if (pin >= 14) pin -= 14; // allow for channel or pin numbers
    ADMUX = ADMUX_VAL;  //(DEFAULT << 6) | (pin & 0x07);
    sbi(ADCSRA, ADSC);
}
// ждем готовности и считываем данные ранее запущенного процесса АЦП
int myAnalogReadGet(){
    uint8_t low, high;
    // ADSC is cleared when the conversion finishes
    while (bit_is_set(ADCSRA, ADSC));
    low  = ADCL;
    high = ADCH;
    return (high << 8) | low; // combine the two bytes
}

#endif

#ifndef STAGE_CODER_FOLDING
String encodeStageProgram(){
    // кодирует стадию в строку, если стадия изменилась отправляет её в порт и eeprom
    program.stageProgram=String( stage.progLine);
    program.stageProgram+=char_SEMI;
    program.stageProgram+=stage.name;
    switch( stage.name ){
        case 'M':
            if(stage.statePulse>0 ){
                if(stage.statePulse == ST__PULSE_DN ) 
                    program.stageProgram+='D';
                else
                    program.stageProgram+='U';
                program.stageProgram+=String(stage.pulseLength);
                program.stageProgram+=',';
                program.stageProgram+=String(stage.pulsePeriod);
                program.stageProgram+=';';
            }        
            if( stage.stateCurrent>0){
                if(stage.stateCurrent==ST__DC_CHARGE) 
                    program.stageProgram+='I';
                else
                    program.stageProgram+='O';
                program.stageProgram += String( stage.currentPWM );
                program.stageProgram +=',';
                program.stageProgram += doubleToString(stage.holdAmps);
                program.stageProgram +=';';
            }        
            program.stageProgram+='V';
            program.stageProgram+=doubleToString(stage.minVoltage);
            program.stageProgram+=',';
            program.stageProgram+=doubleToString(stage.maxVoltage);
            program.stageProgram+=';';
            program.stageProgram+='T';
            program.stageProgram+=doubleToString(stage.tg);
            program.stageProgram+=',';
            program.stageProgram+=intToString(stage.tgTime,0);
            break;
        case 'W':
            program.stageProgram+=intToString(stage.waitTime, 0);
            break;
        case '?':
            if( stage.minVoltage>0 ){
                program.stageProgram+='<';
                program.stageProgram+=doubleToString(stage.minVoltage);
            }                
            if(stage.maxVoltage>0){
                program.stageProgram+='>';
                program.stageProgram+=doubleToString(stage.maxVoltage);
            }
    }            
    if( stage.nextProgLine ){
        if( stage.name != 'G'){
            program.stageProgram+=';';
            program.stageProgram+='G';
        }            
        program.stageProgram+=intToString(stage.nextProgLine, 0);
    }
    return program.stageProgram;
}

#ifdef DEBUG_PROGRAM
void debugStage(){
    Serial.println( logM+ intToString( stage.progLine, 0)+char_SEMI + stage.name);
    switch( stage.name ){
        case 'M':
            if( stage.statePulse > 0 ){
                if( stage.statePulse == ST__PULSE_UP ) 
                    Serial.print( logM+"U" );
                else
                    Serial.print( logM+"D" );
                Serial.print( stage.pulseLength );
                Serial.print( ',' );
                Serial.println( stage.pulsePeriod );
            }
            if( stage.stateCurrent > 0 ){
                if( stage.stateCurrent == ST__DC_CHARGE )
                    Serial.print( logM+"I" );
                else
                    Serial.print( logM+"O" );
                Serial.print( stage.currentPWM );
                Serial.print( ',' );
                Serial.println( stage.holdAmps );
            }
            Serial.print( logM+"V" );
            Serial.print( stage.minVoltage );
            Serial.print( ',' );
            Serial.println( stage.maxVoltage );

            Serial.print( logM+"T" );
            Serial.print( stage.tg );
            Serial.print( ',' );
            Serial.println( stage.tgTime );
            break;
        case 'W':
            Serial.println( stage.waitTime);
            break;
        case '?':
            Serial.print( stage.minVoltage );
            Serial.print( ',' );
            Serial.println( stage.maxVoltage );
            break;
    }
    Serial.print( logM+';');
    Serial.println( stage.nextProgLine );
}
#endif

bool decodeStageProgram( int stageIndex=-1 ){
//    обнуляем стадию
    memset( &stage, 0, sizeof(stage) );
    String sp=program.stageProgram;
    if( stageIndex >=0 ){
        program.stageIndex=stageIndex;
        program.stageProgram=sp=eeStringRead( program.eeStages[ program.stageIndex ] );
    }
    if( sp.length()==0 ) return false;
    int8_t i=sp.length()-1;
    if( sp.charAt( i )==';' ) sp.remove( i );    
    if( sp.length()==0 ) return false;

    i=sp.indexOf(':');
    if( i >= 0 ){ // если вначале строки есть номер, вырезаем его
                  // его не будет если программа набита кнопками. похоже
        stage.progLine=sp.toInt();
        sp.remove(0,i+1);
    }
    // в первом символе строки закодировано действие выполняемое стадией
    char stageName=sp.charAt(0);  // в стадию пока не пишем, потм, если всё остальное сработает
    sp.remove(0,1);

    String subStage;
    switch( stageName ){ 
        case 'M':{ // зарядка/разрядка 10:MU50,500;V11.8,13.600;D5;0.0;T-0.02,60;L20;
            do{
                if( (i=sp.indexOf(';'))>0 ){
                    subStage=sp.substring(0,i);
                    sp.remove(0,i+1);
                }else{
                    subStage=sp;
//                    sp.remove(0);
                    sp="";
                }
                if( !subStage.length() )break;
                char subName=subStage.charAt(0);
                subStage.remove(0,1);
                switch( subName ){
                    case 'U': 
                    case 'D': 
                        if( subName == 'U' )
                            stage.statePulse=ST__PULSE_UP; 
                        else
                            stage.statePulse=ST__PULSE_DN; 
                        stage.pulseLength=subStage.toInt();
                        subStage.remove( 0, subStage.indexOf(",")+1 );
                        stage.pulsePeriod=subStage.toInt();
                        break;
                    case 'I':
                    case 'O':
                        if( subName== 'I' ) 
                            stage.stateCurrent = ST__DC_CHARGE; 
                        else
                            stage.stateCurrent = ST__DC_DISCHA; 
                        stage.currentPWM = subStage.toInt();
                        subStage.remove( 0, subStage.indexOf(",")+1 );
                        stage.holdAmps = subStage.toFloat();
                        break;
                    case 'V':
                        stage.minVoltage=subStage.toFloat();
                        subStage.remove( 0, subStage.indexOf(",")+1 );
                        stage.maxVoltage=subStage.toFloat();
                        break;
                    case 'T':
                        stage.tg=subStage.toFloat();
                        subStage.remove( 0, subStage.indexOf(",")+1 );
                        stage.tgTime=subStage.toInt();
                        break;
                    case 'G':
                        stage.nextProgLine=subStage.toInt();
                        break;
                    default:
                        return false;
                }
            }while( sp.length()>0 );
       }            
       case 'W':{ // ожидание W3600;G10
           stage.waitTime=sp.toInt();
           break;
       }
       case '?':{ // проверка напряжения ?>12.0;G10
           char checkFor=sp.charAt(0);
           sp.remove(0,1);
//           memset(&stage.minVoltage, 0, sizeof(stage.minVoltage)*2);
           if(checkFor=='>') stage.maxVoltage= sp.toFloat();
           if(checkFor=='<') stage.minVoltage=sp.toFloat();
           if( stage.minVoltage<=0 && stage.maxVoltage<=0 ) return false;
           break;
       }
       case 'G': 
            stage.nextProgLine = sp.toInt();
            break;
       default:
            return false;
    }
//    после последней ; может быть номер следующей стадии в программе
    if( stage.nextProgLine==0 && (i=sp.indexOf('G'))>=0 ){
        sp.remove( 0, sp.indexOf('G')+1 );
        stage.nextProgLine = sp.toInt();
    }    
    stage.name=stageName;
#ifdef DEBUG_PROGRAM
debugStage();
#endif;
    return true;   
}
#endif

double calcCurrent(){
// ожидаемый результат относительно АКБ, >0 - заряд, <0 разряд
    double pulseCurrent=0, dcCurrent=0;
    if( stage.statePulse ){
        pulseCurrent  = ( KOEFF_MAIN_COIL * stage.pulseLength )*0.5; // среднее значение для треугольника
        pulseCurrent /= ( stage.pulsePeriod/stage.pulseLength );     // делим на скважность
        if( stage.statePulse == ST__PULSE_DN ) pulseCurrent = pulseCurrent*-1;
    }
    if( stage.stateCurrent==ST__DC_CHARGE ){
        dcCurrent = KOEFF_CHARGE_PWM * stage.currentPWM;
    }        
    if( stage.stateCurrent==ST__DC_DISCHA ){
        dcCurrent =-KOEFF_DISCHARGE_PWM * stage.currentPWM;
    }
    return (pulseCurrent + dcCurrent);
}

#ifndef LCD_FOLDING
// ========================== LCD
void lcdWriteRow( byte Y, String s, uint8_t startX=0){
    if( keySelectCount>0 ){
        if( Y==0 ){
            s= progMemToString(str_CANCEL);
            if( state & ST__PROGRAMM ) 
                s+=progMemToString(str_PROG);
            else
                s+=progMemToString(str_STAGE);
        }else        
            s=String( 3-keySelectCount );
    }        
    lcd.setCursor(0, Y);
    byte j=s.length();
    for( byte i=startX; i<startX+16; i++ ) lcd.print( (i<j?s.charAt(i):' ') );

    if( state>0 ){
        if( Y==0   ){
            lcd.setCursor(15, 0);
            lcd.print( ch_happy );
        }else{
            lcd.setCursor(14, 1);
            switch( stage.statePulse){
                case 0: lcd.print( ch_scul ); break;
                case ST__PULSE_UP: lcd.print( ch_arrow_up );break;
                case ST__PULSE_DN: lcd.print( ch_arrow_dn );break;
            }
            switch( stage.stateCurrent){
                case 0: lcd.print( ch_scul );break;
                case ST__DC_CHARGE: lcd.print( ch_arrow_up );break;
                case ST__DC_DISCHA: lcd.print( ch_arrow_dn );break;
            }
        }            
    }
}
void calcSensorsLCD(){
    nowV_LCD = (double)sensorsDataLCD[SENS_VOLTAGE]/sensorsDataLCD[SAMPLE_NUM];
    nowV_LCD = nowV_LCD*SENSORS_MAX[1] * ONE_DIV_1024;
    double I =(double)sensorsDataLCD[SENS_CURRENT]/sensorsDataLCD[SAMPLE_NUM]-512;
    I = -(SENSORS_MAX[3]*ONE_DIV_1024 * I + DC_CURRENT_SENSOR_ERROR/1000.0);
    memset( sensorsDataLCD, 0, sizeof(sensorsDataLCD) );
    String s= doubleToString(nowV_LCD, 0, 2 );
    s+=progMemToString(str_V);
    s+=doubleToString(I, 0, 2 );
    s+=progMemToString(str_A);
    lcdWriteRow( 1, s );
}
void pulseCurrent_row0( const bool pulse=0 ){
    String s=doubleToStringStepped( calcCurrent() );
    if( s.charAt(0) != '-')  s="+"+s;
    s=progMemToString( (pulse==0?str_PULSE:str_DC) )+s;
    s+=progMemToString( str_A );
    lcdWriteRow( 0, s );    
}
void lcdStatus(){
        #ifdef DEBUG
            Serial.println( String("lcdStatus() currCommand=")+String( int(nowOnScreen))+" "+progMemToString( commands, nowOnScreen ) );
        #endif

    if( state>0 && getUnixTime()-keyPressUnixTime > 10){ // в "стадии" принудительный возврат на экран статуса через 10 сек
        nowOnScreen = VIEW__STAGE;
    }

// первая строка    
    switch( nowOnScreen ){
        case VIEW__CALIBRATOR: lcdWriteRow(0, progMemToString(commands, nowOnScreen) ); break;
        case VIEW__PULSE     : pulseCurrent_row0(); break;
        case VIEW__CURRENT   : pulseCurrent_row0(1); break;
        default:              lcdWriteRow(0, progMemToString(commands, nowOnScreen) ); 
    }        
        
// вторая строка

    switch( nowOnScreen ){
        case VIEW__PULSE:
            switch( pagePulse ){
                case PG_PULSE_DIR:
                    if( stage.statePulse == ST__PULSE_UP )
                        lcdWriteRow(1, progMemToString(str_CHARGE));
                    else
                    if( stage.statePulse == ST__PULSE_DN )
                        lcdWriteRow(1, progMemToString(str_DISCHARGE));
                    else
                        lcdWriteRow(1, progMemToString(str_NONE));
                break;
                case PG_PULSE_WIDTH:
                    lcdWriteRow(1, progMemToString(str_PULSE)+" "+intToString(stage.pulseLength) );
                    break;
                case PG_PULSE_PERIOD: 
                    lcdWriteRow( 1, progMemToString(str_PERIOD)+" "+String(stage.pulsePeriod) );
                    break;
            }                
        break;        
        case VIEW__CURRENT:
            switch( pageCurrent){
                case PG_DC_DIR:
                    if( stage.stateCurrent == ST__DC_CHARGE )
                        lcdWriteRow(1, progMemToString(str_CHARGE));
                    else
                    if( stage.stateCurrent == ST__DC_DISCHA )
                        lcdWriteRow(1, progMemToString(str_DISCHARGE));
                    else
                        lcdWriteRow(1, progMemToString(str_NONE));
                break;
                case PG_DC_PWM:
                     lcdWriteRow(1, progMemToString(str_PWM)+intToString(stage.currentPWM) );
                break;   
            }
            break;                
        case VIEW__HOLDS:{
            switch( pageHolds ){
                case PG_HOLD_AMPS:
                    lcdWriteRow(1, progMemToString(str_HOLD_AMPS)+" "+ doubleToString(stage.holdAmps)+progMemToString(str_A));
                    break;
                case PG_HOLD_MIN_VOLTS:
                    lcdWriteRow(1, progMemToString(str_HOLD_MIN_VOLTS)+" "+ doubleToString(stage.minVoltage)+progMemToString(str_V));
                    break;
                case PG_HOLD_MAX_VOLTS:
                    lcdWriteRow(1, progMemToString(str_HOLD_MAX_VOLTS)+" "+ doubleToString(stage.maxVoltage)+progMemToString(str_V));
                    break;
            }
            break;
        }
        case VIEW__STAGE:{
            lcdWriteRow(1, encodeStageProgram(), stageProgramDisplayX);
            break; 
        }
        default:;
           lcdWriteRow(1, " " );
    }
};
#endif
//=========== установка состояния импульсов и пост тока, глупые состояния не контролируются
void setState( int8_t stPulse=-1, int8_t stCurrent=-1, bool force=false ){ // отрицательное не обрабатываем
    if( stPulse >= 0 && ( force || stPulse != stage.statePulse ) ){
        if( stPulse ){
            if( digitalRead(PIN_PULSE_UP_DN) != !(bool)( (stPulse & ST__PULSE_DN) > 0 ) ){ // будет ли смена состояния реле?
                digitalWrite( PIN_PULSE_UP_DN,  !(bool)( (stPulse & ST__PULSE_DN) > 0 ) );
                delay(RELAY_WAIT); // ждем пока контакты реле долетят
            }
            analogWrite( PIN_PULSE, 1 );   // ардуина сделает все нужные проверки и включит ШИМ
            initPulsePWM( stage.pulsePeriod, stage.pulseLength ); // теперь выставляем нужный нам режим
        }else{
            analogWrite( PIN_PULSE, 0 );           // выключаем шим
            if( digitalRead(PIN_PULSE_UP_DN) == LOW ){// сейчас реле включено
                delay(RELAY_WAIT); // переходные процессы ( монашка )
                digitalWrite( PIN_PULSE_UP_DN, HIGH ); // отпускаю реле
                delay(RELAY_WAIT); // оно щелкает
            }
        }
        stage.statePulse = stPulse;
    }
    if( stCurrent >=0 && ( force || stage.stateCurrent != stCurrent ) ){
        if( stCurrent > 0 ){
            analogWrite( PIN_CHA_DIS_PWM, stage.currentPWM );
            digitalWrite( PIN_CURRENT_CHA_DIS, (bool)( (stCurrent & ST__DC_CHARGE) > 0 ) );
        }else{
            analogWrite( PIN_CHA_DIS_PWM, 0 );
            digitalWrite( PIN_CURRENT_CHA_DIS, HIGH );// вЫключит реле
        }
        stage.stateCurrent = stCurrent;
    }
}

#ifndef DATALOG_FOLDING
void dataLogWrite(String s){
    if(program.programming) return; // на время программирования
    Serial.println( s );
}
/*
void dataLogHeader(){
    String s = progMemToString( str_HEADER0 );
    s.replace("VAL", String( SENSORS_MAX[ 1 ] ));
#ifndef DEBUG_DISABLE_DATALOG 
    dataLogWrite( s );
#endif    

    s = progMemToString( str_HEADER1 );
    s.replace("VAL", String( SENSORS_MAX[ 2 ]));
#ifndef DEBUG_DISABLE_DATALOG
    dataLogWrite( s );
#endif

    s = progMemToString(str_HEADER2); // #text
    s.replace("OPTS", String( OSCILL_POINTS ) );
#ifndef DEBUG_DISABLE_DATALOG
    dataLogWrite( s );
#endif

    s = progMemToString(str_HEADER3); // #oscill
#ifndef DEBUG_DISABLE_DATALOG
    dataLogWrite( s );
#endif
    s = progMemToString(str_HEADER4); // #sysMessagesDone
#ifndef DEBUG_DISABLE_DATALOG
    dataLogWrite( s );
#endif
    lastLogHeaderWriting=millis();
}
void dataLogBegin(){
    dataLogHeader();
#ifndef DEBUG_DISABLE_DATALOG
    dataLogWrite("p:"+encodeStageProgram());

#endif
}
*/
void dataLogWriteDataLCD( bool keepOrigin=false ){
    // отчёт на LCD, динамика рулит!
    uint32_t nowUnixTime= getUnixTime();
    String logData="";

    nowV = (double)sensorsData[SENS_VOLTAGE]/sensorsData[SAMPLE_NUM];
    nowV =  nowV*SENSORS_MAX[1] * ONE_DIV_1024;

    nowI = (double)sensorsData[SENS_CURRENT]/sensorsData[SAMPLE_NUM]-512;
    nowI = -(nowI*SENSORS_MAX[3] * ONE_DIV_1024+ DC_CURRENT_SENSOR_ERROR/1000.0);

    if( state>0 ){
        if( nowOnScreen == VIEW__STAGE ){
            switch( stage.name ){
                 case 'W':
                    logData=progMemToString(str_WAIT);
                    nowUnixTime = stage.waitingFor-nowUnixTime;
                    break;
                 default:
                    nowUnixTime -= stage.time;
            }
            String s=toHours(nowUnixTime, char_SEMI);
            if( nowUnixTime<3600 ) 
                s.remove(0,1);               // часа нет, отрезаем начало
            else
                s.remove(s.lastIndexOf(char_SEMI));// час есть, отрезаем ))
            
            logData += s;
            
            if(stage.name=='M'){
                if( stage.watts>0 )  logData+='+';
                logData+=doubleToStringStepped( stage.watts );
                if( stage.amps>0 ) logData+='+';
                logData+=doubleToStringStepped( stage.amps );
            }                
            lcdWriteRow(0, logData );
        }else{
            if( nowUnixTime-keyPressUnixTime>10 ){ // возврат с любого экрана на "стадию" через 10 сек после последнего нажатия кнопки
                nowOnScreen = VIEW__STAGE;
            }
        }            
    }
}
void dataLogWriteData(){
    dataLogWriteDataLCD();
    uint16_t nowUnixTime= getUnixTime();
    String logData;
    logData = "g:";
    logData+=doubleToString(nowV, 0, 3); // 0 напряжение
    logData+=char_SEMI;
    logData+=doubleToString(nowI, 0, 3); // 1 ток
    logData+=char_SEMI;
    // считаем мгновенную мощность U*I и добавляем в строку
    double momentWatt= nowV * nowI;
    if(stage.name=='M' && lastSampleTime>0 ){
        double timeMult=((nowUnixTime-lastSampleTime )*(1/3600.0));
        stage.watts += ((momentWatt + prevSampleWatt)*0.5) * timeMult;
        stage.amps  += ((nowI + prevSampleAmps )     *0.5) * timeMult;
    }
    
    lastSampleTime=nowUnixTime;
    prevSampleWatt=momentWatt;
    prevSampleAmps=nowI;
    if( state > 0 )
        logData+= doubleToString( stage.amps, 0, 2 );     // 2 ёмкости стадии ач
    logData+= char_SEMI;
    if( state > 0 )
        logData+= doubleToString( stage.watts, 0, 2 );    // 3 ёмкости стадии вч
    logData+= char_SEMI;
    if( state > 0 )
        logData+= toHours( nowUnixTime-stage.time, '|' ); // 4 время работы стадии
    logData+= char_SEMI;
    if( state & ST__PROGRAMM > 0)
        logData+= toHours( nowUnixTime-program.time, '|');// 5 время работы программы
    logData+= char_SEMI;
    if( state & ST__PROGRAMM > 0)
        logData+= doubleToString( program.amps, 0, 2 );   // 6 ёмкости программы ач
    logData+= char_SEMI;
    if( state & ST__PROGRAMM > 0)
        logData+= doubleToString( program.watts, 0, 2 );  // 7 ёмкости программы вч
    logData+= char_SEMI;
    logData+= intToString( freeRam(), 0 );                // 8 свободная память
    
    // пробуем записать данные в лог
    #ifndef DEBUG_DISABLE_DATALOG    
        dataLogWrite(logData);
    #endif    
    if( stage.statePulse ){ // если пульсируем, снимаем данные осцилла, добавляем в лог
        logData = "o";// есои нет, stageEnd() отправит гашение
        for( byte i=OSCILL_START-1; i<(OSCILL_START+OSCILL_POINTS); i++){
            sensorsData[ i ]/=sensorsData[SAMPLE_NUM];
            logData+= char_SEMI;
            logData+= String(sensorsData[ i ]);
        }
        // пробуем записать данные в лог
        #ifndef DEBUG_DISABLE_DATALOG
        dataLogWrite(logData);
        #endif
    }// осцилл гасится в stageEnd
    #ifndef DEBUG_DISABLE_DATALOG
        Serial.flush();
    #endif
}
#endif

#ifndef COMMANDS_KEYS_FOLDING
//==========================
void cmdSelect( uint8_t inc=1 ){ // параметр чтобы отличать вызов от кнопки(==1) или из cmdValueAdd (==0)
    switch( nowOnScreen ){
        case VIEW__PULSE    : 
            pagePulse+=inc;
            if(pagePulse>PG_PULSE_PERIOD) pagePulse=PG_PULSE_DIR;
            if(state>0 && pagePulse==PG_PULSE_DIR ) pagePulse++;
            break;
        case VIEW__CURRENT  :
            if( state>0 ){
                pageCurrent=PG_DC_PWM;
            }else{
                pageCurrent+=inc;
                if(pageCurrent>PG_DC_PWM) pageCurrent=PG_DC_DIR;
            } 
                           
            break;
        case VIEW__HOLDS    :
            pageHolds+=inc;
            if(pageHolds>PG_HOLD_MAX_VOLTS) pageHolds=PG_HOLD_AMPS;
            break;
        case VIEW__STAGE :
            if( inc && state==0 ) stageBegin();
            break;
        case VIEW__PROGRAMM :
            if( inc && state==0 ) programmBegin( -1, 0);
            break;
    }
    // прерывание стадии
    if( inc && state>0 && (nowOnScreen==VIEW__STAGE || nowOnScreen==VIEW__PROGRAMM ) && ( getUnixTime()-stage.time > 10 )){
        if( keySelectCount>2 ){
            state &= ~ST__PROGRAMM;
            stageEnd();
            programEnd();
            state=0; // сброс программы если работает
            nowOnScreen=VIEW__CALIBRATOR;
        }else{
            keySelectCount++;
            lcdWriteRow(0,""); // cansel stage/prog
            lcdWriteRow(1,""); // 3-keySelectCount 
            return;
        }
    }
    lcdStatus();
}
void cmdValueAdd( char inc ){
    cmdSelect(0); // выправит страницы для режима state>0
    switch( nowOnScreen ){
        case VIEW__PULSE:{
            switch( pagePulse ){
                case PG_PULSE_DIR:{
                    int8_t st=stage.statePulse+inc;
                    if(st>2)st=0;
                    if(st<0)st=2;
                    setState(st);
                    break;
                }
                case PG_PULSE_WIDTH:{
                    stage.pulseLength += inc;
                    stage.pulseLength = max(10, min(stage.pulsePeriod-10, stage.pulseLength ));
                    eeWordWrite( ADDR_PULSE_LENGTH_0, stage.pulseLength );
                    initPulsePWM( stage.pulsePeriod,  stage.pulseLength );
                    break;
                }
                case PG_PULSE_PERIOD:{
                    stage.pulsePeriod+= 10*inc;
                    stage.pulsePeriod=max(stage.pulseLength+10, min(8000,stage.pulsePeriod));
                    stage.pulsePeriod = int(stage.pulsePeriod/10)*10;
                    eeWordWrite( ADDR_PULSE_PERIOD_0, stage.pulsePeriod);
                    initPulsePWM( stage.pulsePeriod,  stage.pulseLength );
                    break;
                }
            }
            break;
        }            
        case VIEW__CURRENT:{
            switch( pageCurrent ){
                case PG_DC_DIR:{
                    int8_t st=stage.stateCurrent+inc;
                    if(st>2)st=0;
                    if(st<0)st=2;
                    setState(-1,st);
                    break;
                }
                case PG_DC_PWM:
                    stage.currentPWM = max(0, min(255, (int)inc + stage.currentPWM ) );
                    if(stage.stateCurrent) analogWrite( PIN_CHA_DIS_PWM, stage.currentPWM);
                    EEPROM.write( ADDR_currentPWM, stage.currentPWM);
                    break;
            }
            break;
        }
        case VIEW__HOLDS:{
            switch( pageHolds ){
                case PG_HOLD_AMPS     : stage.holdAmps  =max(0, stage.holdAmps  + inc*0.01 ); break;
                case PG_HOLD_MIN_VOLTS: stage.minVoltage=max(0, stage.minVoltage+ inc*0.01 ); break;
                case PG_HOLD_MAX_VOLTS: stage.maxVoltage=max(0, stage.maxVoltage+ inc*0.01 ); break;
            }
            break;
        }
        case VIEW__STAGE:{
            dataLogWriteDataLCD( true );
            if( state>0 ) return;
            stageProgramDisplayX+=inc;
            stageProgramDisplayX=max( 0, stageProgramDisplayX);
            stageProgramDisplayX=min(program.stageProgram.length()-18,  stageProgramDisplayX);
            break;
        }
    }                
    if( state>0 ) encodeStageProgram(); // зекодирует и запишет в eeprom
    lcdStatus();
}
void cmdAdd( char inc){
    nowOnScreen += inc;
    int8_t minCmd;
    if( state>0 ){
        minCmd= VIEW__PULSE;
        if(stage.stateCurrent == 0 && nowOnScreen == VIEW__CURRENT ) nowOnScreen += inc;
        if(stage.statePulse   == 0 && nowOnScreen == VIEW__PULSE   ) nowOnScreen += inc;
    }else{
        minCmd= 0;
    }
    if( nowOnScreen>VIEW__LAST_CMD) nowOnScreen=minCmd;
    if( nowOnScreen<minCmd ) nowOnScreen=VIEW__LAST_CMD;
    cmdValueAdd(0);
}
//==========================
byte getKey(){
    byte k;
    uint8_t oADMUX=ADMUX;
//    myAnalogReadStart( ADMUX_LCD_KEYS );// иначе фальшивые срабатывания на кнопках близких к последнему измеренному напряжению
    ADMUX=ADMUX_LCD_KEYS;
    sbi(ADCSRA, ADSC);
    myAnalogReadGet();
    sbi(ADCSRA, ADSC);
    uint16_t input=myAnalogReadGet();
    ADMUX=oADMUX;
    for (k = 0; k < NUM_KEYS; k++){ // convert into key press
      if (input <= adc_key_val[k]) return k;
    }
    return NUM_KEYS;     // No valid key pressed
}
void handleKey(){
    switch( key ){
        case KEY_RIGHT : cmdValueAdd( 1); break;
        case KEY_UP    : cmdAdd(1); break;
        case KEY_DOWN  : cmdAdd(-1); break;
        case KEY_LEFT  : cmdValueAdd(-1); break;
        case KEY_SELECT: cmdSelect();
    }
}
bool doKeys(){
    key = getKey();
    if (key >= NUM_KEYS ){
        oldKey = key;
        return false;
    } 
        
    uint32_t Millis=millis();
    if (key != oldKey){       // сменилось состояние клавишь
        delay(50);            // проверяем дребезг
        key = getKey();
        if (key != oldKey){
            keyPressMills=Millis; // время последней смены клавиши
            keyPressDiff =1000;   // пауза до начала повторов, если будет удержано дольше, повторы пойдут
            keyPressRepeat=0;     // сброс счётчика повторений
            if( key == KEY_SELECT ){
                keyPressMills=0; // чтоб небыло повтора "SELECT"
                if( getUnixTime()-keyPressUnixTime>5 ) keySelectCount=0;
            }else{
                keySelectCount=0; // для кнопки "SELECT" считаем к-во последовательных нажатий чтоб остановить программу после трёх
            }                
            oldKey = key;         // фиксируем теперешнюю кнопку как последнюю нажатую
            handleKey();          // обрабатываем нажатие
            keyPressUnixTime=getUnixTime(); // фиксируем время нажатия, для возврата на экран программы после 10 сек неактивности кнопкодава
            return true;          // ура!
        }            
    }else{// клавиша не сменилась, удержание
        if( keyPressMills && (Millis-keyPressMills) > keyPressDiff ){ // прошло достаточно времени для автоповтора?
            keyPressRepeat++;                                           // счётчик автоповторов
            keyPressDiff  = keyPressRepeat < 5 ? 500 :                  // после того как автоповторили > 5 раз меняем скорость 
                                (keyPressRepeat < 10 ? 200 : 100) ;     // после 10 раз меняем скорость ещё раз
            keyPressMills = Millis;                                     // фиксируем время
//            oldKey = key;
            handleKey();
            keyPressUnixTime=getUnixTime();                             // фиксируем время нажатия, для возврата на экран программы после 10 сек неактивности кнопкодава
            return true;                                                // ура!
        }            
    }
    return false; // нет кнопки
}
#endif

#ifndef STAGE_FOLDING

bool stageQueryNext( uint16_t nextLine=0 ){
    if( nextLine==0 ) nextLine = stage.nextProgLine;
    if( nextLine > 0 ){ // есть указание на след строку
        for(int i=0; i<program.progLinesCount; i++){
            if( program.indexToLine[i]==nextLine ){
                program.stageIndex=i; // если стадия не расшифруется, сюда нужно будет -1 загнать
                program.stageProgram=eeStringRead( program.eeStages[program.stageIndex] );
                #ifdef DEBUG
                    Serial.println(program.stageProgram);
                #endif
                return true;
            }
        }
    }else{
        nextLine= program.stageIndex >= 0 ? program.stageIndex+1 : 0;
        while( nextLine < MAX_PROG_LINES ){ // c 0 начинается
            program.stageIndex=nextLine; // если сталия не расшифруется, сюда нужно будет -1 загнать
            program.stageProgram=eeStringRead( program.eeStages[program.stageIndex] );
            if( program.stageProgram.length()>0){
                #ifdef DEBUG
                    Serial.println(program.stageProgram);
                #endif
                return true;
            }                
            nextLine++;
        }
    }
    program.stageProgram="";
    #ifdef DEBUG
        Serial.println();
    #endif
    program.stageIndex=-1;
    return false;
}
void stageBegin(){
    lcdWriteRow(1, progMemToString(str_STARTING));
    prevSampleWatt      =0;
    lastLCDshowTime     =0;
    lastLogWriting      =0;

    stage.time= getUnixTime();
    if( stage.name=='W' ) stage.waitingFor = stage.waitTime + stage.time;
    state |= ( calcCurrent()>0 ? ST__STAGE_CHARGE : ST__STAGE_DISCHARGE );
    nowOnScreen=VIEW__STAGE;          // чего экран будет показывать
    setState(stage.statePulse, stage.stateCurrent, 1); // запуск PWM и сигнальных ног
    if( !stage.statePulse ) dataLogWrite("o:");// погасит линию на осцилле
    
    String s="b:";
    s+=program.stageIndex;
    s+=char_SPACE;
    s+=encodeStageProgram();
    dataLogWrite(s);
    if( stage.name=='M'){
        digitalWrite(PIN_POWER_ON, HIGH);
        delay(RELAY_WAIT);
    }    
    if( stage.stateCurrent>0 )
        digitalWrite( PIN_DC_POWER_ON, LOW ); // включит реле
};
void stageEnd(){
    digitalWrite( PIN_DC_POWER_ON, HIGH );    // вЫключит реле
    if( digitalRead(PIN_POWER_ON) ){
        digitalWrite( PIN_POWER_ON, LOW ); // это реле включает +5
        delay( RELAY_WAIT );
    }  
    state = state & ST__PROGRAMM;
    lcdStatus();
    stage.time=getUnixTime()-stage.time;
    if( program.stageIndex >=0 ){
        program.stats[ program.stageIndex ].execCount++;
        program.stats[ program.stageIndex ].time += stage.time;
        program.stats[ program.stageIndex ].watts+=stage.watts;
        program.stats[ program.stageIndex ].amps +=stage.amps;
        program.watts+= stage.watts;
        program.amps += stage.amps;
        String s="e:";
        s+=stage.progLine;
        s+=char_SEMI;
        s+=stage.name;
        s+=char_SPACE;
//        s+=program.stages[ program.stageIndex ];
//        s+=char_SPACE;
        s+=toHours(stage.time,char_SEMI);
        s+=char_SPACE;
        s+=doubleToString(stage.amps,0,3);
        s+="Ah ";
        s+=doubleToString(stage.watts,0,3);
        s+="Wh";
        dataLogWrite(s); // запись об окончании стадии
    }

    if(state & ST__PROGRAMM){
        if ( stageQueryNext() && decodeStageProgram() ){
            stageBegin(); // есть следующая и загрузилась
        }else{
            programEnd();
        }
    }
    if( !state ){
        nowOnScreen=VIEW__CALIBRATOR;
    }
};
bool stageCheckEnd(){
    switch( stage.name ){
        case 'M': {
            String s="c:";
            s+=doubleToString(stage.minVoltage );
            s+=" <= ";
            s+=doubleToString(nowV);
            s+=" <= ";
            s+=doubleToString( stage.maxVoltage);
            dataLogWrite( s );
            if(  nowV <= stage.minVoltage ){ 
                return true;
            } // контроль напряжения
            if( nowV >= stage.maxVoltage ) { 
                return true; // контроль напряжения
            }                
            break;
        }
        case 'W':{
            if( getUnixTime()>=stage.waitingFor ) {
                return true;
            }                
            break;
        }

        case '?':{
            if( stage.maxVoltage > 0 ){
                 if( nowV < stage.maxVoltage ){ // проверяем на невыполнение
                     stage.nextProgLine=0; //переход к следующей строке будет а не goto
                 }
            }
            if( stage.minVoltage > 0 ){
                 if( nowV > stage.minVoltage  ){
                     stage.nextProgLine=0; //переход к следующей строке будет а не goto
                 }
            }                 
            return true; // выполнена в любом случае
        }
        case 'G':{
            return true;// выполнена в любом случае
        }             
    }        
    return false;
}

void programFillIndexToLine(){
    String s="";
    char ch;
    program.progLinesCount=EEPROM.read( ADDR_EEPROGRAM_LINES_COUNT );
    
    if(program.progLinesCount==0xFF ) program.progLinesCount=0;
    for( uint8_t i=0; i<program.progLinesCount; i++ ){
        uint8_t j=1;
        s="";
        do{
            ch=EEPROM.read( program.eeStages[i]+j );
            if( ch<'0' || ch>'9' ) break;
            s+=ch;
            j++;
        }while( j<4 );
        program.indexToLine[i]=s.toInt();

        if( program.indexToLine[i]==0 ) break;
    }
/*
    Serial.print("m:");
    for(uint8_t i=0; i<program.progLinesCount; i++ ){
        Serial.print(i);
        Serial.print("=>");
        Serial.println(program.indexToLine[i]);
            
    }
    Serial.println("");
*/        
}

void programReadFromEEPROM(){
    program.progLinesCount=EEPROM.read( ADDR_EEPROGRAM_LINES_COUNT );
    if( program.progLinesCount>0 && program.progLinesCount<0xFF ){
        uint8_t i=0;
        program.eeStages[i]=ADDR_EEPROGRAM;
        do{
            serialString=eeStringRead( program.eeStages[i] );
            if( serialString.length()>0 ){
                if( i < program.progLinesCount-1 )
                    program.eeStages[i+1]=program.eeStages[i]+serialString.length()+1;
            }
            i++;
        }while( serialString.length()>0 && i<program.progLinesCount );
        programFillIndexToLine();
    }else{
        program.progLinesCount=0;
    }        
}

void programmBegin( int str, uint8_t ln ){
// запуск программы со строки с индексом str или с номером ln
// у номера ln приоритет
    if( ln>0 ){
        for( int i=0; i<program.progLinesCount; i++){
            if( program.indexToLine[i]==ln ){
                str=i;
                break;
            }
        }
    }
    if( !decodeStageProgram(str) ) return;
    state |= ST__PROGRAMM; // добавление бита "работа по программе", если нужно
    lcd.setCursor(15, 0);  // рожу нарисует, чтоб понятно было
    lcd.print( ch_happy );
    for( int i=0; i<MAX_PROG_LINES; i++)
        memset( &program.stats[i], 0, sizeof(program.stats[i]) );
    program.amps = program.watts=0;
    stageBegin();
    program.time = stage.time;
}

void programEnd(){
// перед вызовом нужно звать stageEnd    
// чтоб рекурсии не было
    state=0; // останов
    program.time = getUnixTime()-program.time;
    program.watts=0;
    program.amps=0;
    for( int i=0; i < program.progLinesCount; i++){
        program.watts+=program.stats[i].watts;
        program.amps +=program.stats[i].amps;
    }
    String s="s:";
    s+=toHours(program.time, ':');
    s+=char_SPACE;
    s+=doubleToString(program.amps,0,3);
    s+="Ah ";
    s+=doubleToString(program.watts,0,3);
    s+="Wh";
    dataLogWrite( s );// запись об окончании программы
    decodeStageProgram(0);
}
#endif

#ifndef ADC_FOLDING
// прерывание от АЦП
ISR( ADC_vect ){
    uint8_t  low = ADCL;
    uint8_t high = ADCH;
    sensorsData[ OSCILL_INDEX++ ] += ( (high << 8) | low );
    if( OSCILL_INDEX >= OSCILL_START+OSCILL_POINTS ){ // всё считали?
        sensorsData[SAMPLE_NUM]++; // добавляем семпл
        cbi(ADCSRA, ADIE);  //disable Interrupts и останавливаемся
        cbi(ADCSRA, ADATE); // запрещаем триггер источника перезапусков измерений. По этому биту ниже определяется завершенность считывний
    }
}
void readOscillStart(){
    OSCILL_INDEX=OSCILL_START;   // начало данных
    ADMUX = ADMUX_VOLTAGE_PULSE; // выбор канала
    sbi(ADCSRA, ADIE); //разрешаем прерывания
    ADCSRB &= ~7; // выбор источника запуска ADC 3 бита в 0 - freerunning
    sbi(ADCSRA, ADATE);// запуск freerunning
    sbi(ADCSRA, ADSC); // запуск первого ADC
}
#endif

#ifndef SERIAL_PORT_FOLDING
void printOk(){
    Serial.println( progMemToString(str_OK));
    Serial.flush();
}
void logProg(){
    uint16_t progSize=0;
    for( int i=0; i<program.progLinesCount; i++){
        String s=eeStringRead( program.eeStages[i]);
        progSize+=s.length();
        progSize++;
        Serial.print( logM );
        Serial.print(i);
        Serial.print(char_SEMI);
        Serial.println(s);
        Serial.flush();
    }
    Serial.println( logM+String(progSize));
    Serial.println( logM+String(1024-100-progSize));
    Serial.flush();
}
void doSerialString(){
    if( serialString.length() == 0 ) return;
    switch( serialString.charAt(0) ){
        case 'h':// комп переподключился к порту
            Serial.println("##"); 
            Serial.flush();
            break;
        case 'p':// старт программы
            if(state==0){
                serialString.remove(0,1);
                uint16_t ln=serialString.toInt();
                programmBegin( 0, ln );
            }                 
            break;
        case 'g':// старт стадии, должен быть ИНДЕКС строки стадии 0,1,2...
            if(state==0){
                serialString.remove(0,1);
                decodeStageProgram( serialString.toInt() );
                stageBegin();
            }                
            break;
        case '#': // останов работающей программы
            state &= ~ST__PROGRAMM; // чтоб stageEnd(); не запустил следующую
            stageEnd();
            programEnd();
            printOk();       
            break;
        case '^': // сброс программы перед заливкой
            if( state==0 ){
                memset( program.eeStages,    0, sizeof( program.eeStages    ));
                memset( program.indexToLine, 0, sizeof( program.indexToLine ));
                eeWordWrite( ADDR_EEPROGRAM_LINES_COUNT, 0); // забьём количество строк и длинну нулевой строки
                program.programming=true;
                program.time=program.watts=program.amps=0;
                for(int i=0; i<MAX_PROG_LINES; i++ ){
                    memset( &program.stats[i], 0, sizeof( program.stats[i] ));     
                }
                printOk();
            }                
            break;
        case '$': // завершение заливки
            decodeStageProgram(0);
            printOk();
            program.programming=false;
        // сюда всякое потом
        case '?':
            logProg();
            printOk();
            break;
        default: // 
            uint8_t dogIndex=serialString.indexOf('@');
            if( program.programming &&  dogIndex > 0 ){ // заливка строки программы
                                                        // формат  0@10:MU....  первая цифра порядковый номер строки от 0
                uint8_t ln=serialString.toInt();
                if( ln<MAX_PROG_LINES ){
                    serialString.remove(0, dogIndex+1 );
                    program.stageProgram=serialString;
                    if( decodeStageProgram( -1 ) ){ // декодирует program.stageProgram
                        program.stageProgram=encodeStageProgram();
                        uint16_t lnAddr;
                        if( ln==0 ) 
                            program.eeStages[ ln ]=ADDR_EEPROGRAM;
                        else{
                            program.eeStages[ ln ] = program.eeStages[ ln-1 ]; // адрес предыдущей строки
                            program.eeStages[ ln ]+= EEPROM.read( program.eeStages[ ln ] )+1;
                        }
                        eeStringWrite( program.eeStages[ ln ], program.stageProgram );
                        program.progLinesCount=ln+1;
                        EEPROM.write(ADDR_EEPROGRAM_LINES_COUNT, program.progLinesCount );
                        Serial.println( progMemToString(str_OK)+">"+String(program.eeStages[ ln ])+"<"+program.stageProgram);
                    }else{
                        
                    }
                    Serial.flush();
                }
            }                
    }
}
#endif

void setup(){
    Serial.begin(115200);
    Serial.println();// на случай если в порту чёнить застряло
    Serial.println("##"); // комп получив это напишет себе заголовок
    Serial.flush();
    
    pinMode( PIN_PULSE, OUTPUT);  // ШИМ инициализирует пусть ардуина

    pinMode( PIN_POWER_ON,        OUTPUT);
    digitalWrite( PIN_POWER_ON,      LOW); // включает +5
    pinMode( PIN_DC_POWER_ON,     OUTPUT);
    digitalWrite( PIN_DC_POWER_ON,  HIGH); // включает 0

    pinMode( PIN_PULSE_UP_DN,     OUTPUT); // импульс заряжает 1 / разряжает 0
    digitalWrite( PIN_PULSE_UP_DN, HIGH ); // моя релюха включается землёй

    pinMode( PIN_CHA_DIS_PWM,     OUTPUT); // шим
    pinMode( PIN_CURRENT_CHA_DIS, OUTPUT); // флаг подключения зарядки 0 / разрядки 1 постоянным током

    // скорость ADC: 100=16=13.5мкс, 010=4; делитель 0 и 2 идут 1023
    sbi(ADCSRA,ADPS2) ; // set prescale to 16
    cbi(ADCSRA,ADPS1) ;
    cbi(ADCSRA,ADPS0) ;

    // LCD init
    lcd.createChar(ch_arrow_up, arrow_up); // загрузка самопальных символов; стрелка вверх
    lcd.createChar(ch_arrow_dn, arrow_dn); // стрелка вниз
    lcd.createChar(ch_scul,     scul    ); // череп
    lcd.createChar(ch_happy,   happy    ); // довольная рожа
    lcd.begin(16,2);

    programReadFromEEPROM();      // читаем программу из EEPROM
    if( !decodeStageProgram(0) ){ // программа была? удачное декодирование?
        eeAllRead();              // нет, читаем простые сохранённые параметры
        if( (( (uint8_t)stage.pulseLength & (uint8_t)stage.pulsePeriod  & stage.statePulse & stage.stateCurrent & stage.currentPWM) ==0xFF) ){
            // первый раз пишем в эту область eeprom
            stage.pulseLength       = 50;
            stage.pulsePeriod       =500;
            stage.statePulse        =  0;
            stage.stateCurrent      =  0;
            stage.currentPWM        = 10;
            eeAllWrite();
        }
    }
    lcdStatus(); // вывод чего-то там

    if( stage.name ){ // если ранее стадия считалась нормально, будет её имя, иначе 0
        //        stageBegin( true ); // это сразу запуск стадии, без подтверждений и кнопок
        setState(stage.statePulse, stage.stateCurrent, 1); // запуск PWM и сигнальных ног, POWER_ON не включится, можно тестером потыкаться
    }else{
        setState(stage.statePulse, stage.stateCurrent, 1); // запуск PWM и сигнальных ног, POWER_ON не включится, можно тестером потыкаться
    }
}
void loop(){
    if( program.programming){
        delay(100);
    }else{
        int16_t tmp;
        if( stage.statePulse ){ // импульсы есть, по ним синхронизируемся
            // ШИМ запущен в режиме phase correct, т.е. считает по кольцу: 0->PERIOD_TICKS( stage.pulsePeriod )->0->PERIOD_TICKS( stage.pulsePeriod )->0.....
            // пока счетчик меньше pulseLengthInTicks=PERIOD_TICKS( stage.pulseLength ) - на ноге 1, когда между pulseLength и pulsePeriod - на ноге 0
            while( TCNT1 > pulseLengthInTicks  ); // в паузе PWM
            // пауза кончилась, пошел импульс
            ADMUX=ADMUX_CURRENT;
            delayMicroseconds(8);  // даём зарядиться ёмкости
            sbi(ADCSRA, ADSC);

            tmp=myAnalogReadGet();
            sensorsData   [SENS_CURRENT] += tmp; // снимаем показания внутри импульса, пока заряжается индуктивность, 
            sensorsDataLCD[SENS_CURRENT] += tmp;
            ADMUX = ADMUX_VOLTAGE; // переключаем мультиплексор ЦАП
            delayMicroseconds(8);  // даём зарядиться ёмкости
            sbi(ADCSRA, ADSC);
            tmp=myAnalogReadGet();
            sensorsData   [SENS_VOLTAGE] +=tmp ;  // сработает если длительность импульса > 40мкс, если нет, сдвинет измерения осцилла
            sensorsDataLCD[SENS_VOLTAGE] +=tmp ;
            ADMUX = ADMUX_VOLTAGE_PULSE; // переключаем мультиплексор ЦАП
            while( TCNT1 < pulseLengthInTicks-PERIOD_TICKS(20) ); //  идет импульс, ёмкости ЦАП заряжаются ))
            // выходим из ожидания раньше на 20мкс, так как
            // 13,5мкс - 1 измерение до импульса, +1,5мкс до захвата семпла = 15мкс
            // и 5мкс до горба напряжения на моём девайсе, у тебя может быть не 5
            // точно в горб не попадем, потому как компилятор как-то скомпилирует и будет небольшая задержка
            // но попадаем очень близко
            readOscillStart(); // запуск считывания данных для осцилла
            while( bit_is_set(ADCSRA, ADATE) ); // ждем пока считаются все точки
            sensorsData[ OSCILL_TIME ] += 13.5*OSCILL_POINTS; // по 13,5 мкс
//            stagePulseCount++;                                // сквозной счётчик импульсов текущего состояния
        }else{ // импульсов нет, просто подождём 
            ADMUX = ADMUX_CURRENT;
            delayMicroseconds(300);
            sbi(ADCSRA, ADSC);
            tmp = myAnalogReadGet(); 
            sensorsData[SENS_CURRENT] += tmp;
            sensorsDataLCD[SENS_CURRENT] += tmp;
            ADMUX = ADMUX_VOLTAGE;
            delayMicroseconds(300);
            sbi(ADCSRA, ADSC);
            tmp = myAnalogReadGet();
            sensorsData[SENS_VOLTAGE] += tmp;
            sensorsDataLCD[SENS_VOLTAGE] += tmp;
            sensorsData[SAMPLE_NUM]++;
        }
        sensorsDataLCD[SAMPLE_NUM]++;

        uint32_t millis_ = millis();
        if( nowOnScreen == VIEW__CALIBRATOR || (nowOnScreen == VIEW__STAGE && state>0)){
            if( millis_-lastLCDshowTime>1000  ){ // обновляем вольтметр и амперметр на LCD раз в секунду
                lastLCDshowTime=millis_;
                calcSensorsLCD();
            }
            if( stage.tgTime>0 ){
                stage.tgTimeCounter++;
                if( stage.tgTimeCounter >= stage.tgTime ){
                    stage.tgVoltage     = nowV_LCD;
                    stage.prevTgVoltage = stage.tgVoltage;
                    stage.tgTimeCounter = 0;
                }
            }                
        }        

        if( millis_-lastLogWriting > LOG_WRITE_FREQUENCY ){
            lastLogWriting=millis_;
            dataLogWriteData();
            if( state>0 && stageCheckEnd() ) stageEnd();
            memset( sensorsData, 0, sizeof(sensorsData));
        }
    
        doKeys();
        ADMUX = ADMUX_CURRENT;  // следующее измерение будет отсюда
    }    

    if( Serial.available() ){
        serialString=Serial.readString();
        serialString.replace("\r","");
        serialString.replace("\n","");
        doSerialString();
        serialString="";
    }
    
}

